#include "Rinternals.h"
#include "R_ext/Rdynload.h"

#include <gdbm.h>
#include <sys/stat.h>

#define IS_PROPER_STRING(s) (TYPEOF(s) == STRSXP && LENGTH(s) > 0)

void fatal(char *msg)
{
    error("%s%", "msg");
}

static GDBM_FILE dbopen(SEXP name, int flag)
{
    GDBM_FILE dbf;
    int mode = S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH;

    if (! IS_PROPER_STRING(name))
	error("not a proper database name");
    
    dbf = gdbm_open(CHAR(STRING_ELT(name, 0)), 0, flag, mode, fatal);
    if (dbf == NULL) {
	char *msg = NULL;
	if (gdbm_errno != 0)
	    msg = gdbm_strerror(gdbm_errno);
	else if (errno != 0)
	    msg = strerror(errno);
	if (msg != NULL)
	    error("%s", msg);
	else
	    error("error opening data base");
    }
    return dbf;
}
    
static SEXP dbnew(SEXP name)
{
    GDBM_FILE dbf = dbopen(name, GDBM_NEWDB);
    gdbm_close(dbf);
    return R_NilValue;
}

static SEXP dbinsert(SEXP name, SEXP keystr, SEXP val)
{
    GDBM_FILE dbf;
    datum key, content;
    int ret;

    if (! IS_PROPER_STRING(keystr))
	error("key not a string");
    if (! IS_PROPER_STRING(val))
	error("value must be a string");

    dbf = dbopen(name, GDBM_WRITER);
    key.dsize = LENGTH(STRING_ELT(keystr, 0)) + 1;
    key.dptr = CHAR(STRING_ELT(keystr, 0));
    content.dsize = LENGTH(STRING_ELT(val, 0)) + 1;
    content.dptr = CHAR(STRING_ELT(val, 0));
    ret = gdbm_store(dbf, key, content, GDBM_INSERT);
    gdbm_close(dbf);
    if (ret != 0)
	error("inserting item failed");
    return R_NilValue;
}

static SEXP dbfetch(SEXP name, SEXP keystr)
{
    GDBM_FILE dbf;
    datum key, content;

    if (! IS_PROPER_STRING(keystr))
	error("key not a string");

    dbf = dbopen(name, GDBM_READER);
    key.dsize = LENGTH(STRING_ELT(keystr, 0)) + 1;
    key.dptr = CHAR(STRING_ELT(keystr, 0));
    content = gdbm_fetch(dbf, key);
    gdbm_close(dbf);
    if (content.dptr == NULL)
	return R_NilValue;
    else {
	SEXP val = allocVector(CHARSXP, content.dsize - 1);
	memcpy(CHAR(val), content.dptr, content.dsize - 1);
	val = ScalarString(val);
	free(content.dptr);   
	return val;
    }
}

static SEXP dbexists(SEXP name, SEXP keystr)
{
    GDBM_FILE dbf;
    datum key;
    Rboolean exists;
    SEXP Rf_mkTrue(void), Rf_mkFalse(void);

    if (! IS_PROPER_STRING(keystr))
	error("key not a string");

    dbf = dbopen(name, GDBM_READER);
    key.dsize = LENGTH(STRING_ELT(keystr, 0)) + 1;
    key.dptr = CHAR(STRING_ELT(keystr, 0));
    exists = gdbm_exists(dbf, key);
    gdbm_close(dbf);
    return exists ? Rf_mkTrue() : Rf_mkFalse();
}

static SEXP dbdelete(SEXP name, SEXP keystr)
{
    GDBM_FILE dbf;
    datum key;
    int ret;

    if (! IS_PROPER_STRING(keystr))
	error("key not a string");

    dbf = dbopen(name, GDBM_WRITER);
    key.dsize = LENGTH(STRING_ELT(keystr, 0)) + 1;
    key.dptr = CHAR(STRING_ELT(keystr, 0));
    ret = gdbm_delete(dbf, key);
    gdbm_close(dbf);
    if (ret != 0)
	error("deletion failed");
    return R_NilValue;
}

/**** ought to do unwind-protect to insure closing */
static SEXP dblist(SEXP name)
{
    GDBM_FILE dbf;
    datum key;
    SEXP list = R_NilValue, str, val;
    int i, count = 0;

    dbf = dbopen(name, GDBM_READER);
    key = gdbm_firstkey(dbf);
    while (key.dptr != NULL) {
	datum oldkey = key;
	count++;
	PROTECT(list);
	str = mkChar(key.dptr);
	list = CONS(str, list);
	key = gdbm_nextkey(dbf, key);
	free(oldkey.dptr);
	UNPROTECT(1);
    }
    gdbm_close(dbf);
    
    PROTECT(list);
    val = allocVector(STRSXP, count);
    for (i = 0; i < count; i++, list = CDR(list))
	SET_STRING_ELT(val, i, CAR(list));
    UNPROTECT(1);
    return val;
}

static R_CallMethodDef CallDefs[] = {
    {"dbnew", (DL_FUNC) dbnew, 1},
    {"dbinsert", (DL_FUNC) dbinsert, 3},
    {"dbfetch", (DL_FUNC) dbfetch, 2},
    {"dbexists", (DL_FUNC) dbexists, 2},
    {"dbdelete", (DL_FUNC) dbdelete, 2},
    {"dblist", (DL_FUNC) dblist, 1},
    {NULL}
};

void R_init_gdbm(DllInfo *info)
{
    R_registerRoutines(info, NULL, CallDefs, NULL, 0);
}
